/*
 * #%L
 * EUGene :: EUGene
 * 
 * $Id: ObjectModelTransformerToJava.java 905 2010-05-14 13:44:45Z tchemit $
 * $HeadURL: http://svn.nuiton.org/svn/eugene/tags/eugene-2.0.2/eugene/src/main/java/org/nuiton/eugene/java/ObjectModelTransformerToJava.java $
 * %%
 * Copyright (C) 2004 - 2010 CodeLutin
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 * 
 * You should have received a copy of the GNU General Lesser Public 
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * #L%
 */

package org.nuiton.eugene.java;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.nuiton.eugene.Template;
import org.nuiton.eugene.models.object.ObjectModel;
import org.nuiton.eugene.models.object.ObjectModelAttribute;
import org.nuiton.eugene.models.object.ObjectModelClass;
import org.nuiton.eugene.models.object.ObjectModelClassifier;
import org.nuiton.eugene.models.object.ObjectModelDependency;
import org.nuiton.eugene.models.object.ObjectModelElement;
import org.nuiton.eugene.models.object.ObjectModelEnumeration;
import org.nuiton.eugene.models.object.ObjectModelInterface;
import org.nuiton.eugene.models.object.ObjectModelModifier;
import org.nuiton.eugene.models.object.ObjectModelOperation;
import org.nuiton.eugene.models.object.ObjectModelParameter;
import org.nuiton.eugene.models.object.ObjectModelTransformer;
import org.nuiton.eugene.models.object.ObjectModelType;

import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * Created: 28 oct. 2009
 *
 * @author fdesbois <fdesbois@codelutin.com>
 */
public abstract class ObjectModelTransformerToJava extends ObjectModelTransformer<ObjectModel> {

    private static final Log log =
            LogFactory.getLog(ObjectModelTransformerToJava.class);

    /** internal builder */
    protected JavaBuilder builder;

    private String constantPrefix;

    @Override
    protected Template<ObjectModel> initOutputTemplate() {
        // by default, use a stupid java generator
        return new JavaGenerator();
    }

    @Override
    protected ObjectModel initOutputModel() {
        if (log.isDebugEnabled()) {
            log.debug("inputModelName = " + getModel().getName());
        }
        builder = new JavaBuilder(getModel().getName());
        setConstantPrefix("");
        return builder.getModel();
    }

    @Override
    protected void debugOutputModel() {
        if (log.isDebugEnabled()) {
            log.debug("classes : " + getOutputModel().getClasses().size());
            for (ObjectModelClass clazz : getOutputModel().getClasses()) {
                log.debug("  class : " + clazz.getQualifiedName());
                ImportsManager manager = getOutputModel().getExtension(
                        clazz.getQualifiedName(), ImportsManager.class
                );
                for (String imports :
                        manager.getImports(clazz.getPackageName())) {
                    log.debug("  import : " + imports);
                }
                for (ObjectModelAttribute attribute : clazz.getAttributes()) {
                    log.debug("  attribute : " + attribute.getType() +
                              " " + attribute.getName());
                }
                for (ObjectModelOperation operation : clazz.getOperations()) {
                    log.debug("  operation : " + operation.getReturnType() +
                              " " + operation.getName());
                }
            }
        }
    }

    protected ObjectModelClass createClass(String name, String packageName) {
        return builder.createClass(name, packageName);
    }

    public ObjectModelEnumeration createEnumeration(String name,
                                                    String packageName) {
        return builder.createEnumeration(name, packageName);
    }

    protected ObjectModelClass createAbstractClass(String name,
                                                   String packageName) {
        return builder.createAbstractClass(name, packageName);
    }

    protected ObjectModelInterface createInterface(String name,
                                                   String packageName) {
        return builder.createInterface(name, packageName);
    }

    protected void setSuperClass(ObjectModelClass classifier,
                                 String superclassQualifiedName) {
        builder.setSuperClass(classifier, superclassQualifiedName);
    }

    protected void setSuperClass(ObjectModelClass classifier,
                                 Class<?> superclassQualifiedName) {
        builder.setSuperClass(classifier, superclassQualifiedName.getName());
    }

    protected void addInterface(ObjectModelClassifier classifier,
                                String interfaceQualifiedName) {
        builder.addInterface(classifier, interfaceQualifiedName);
    }

    protected void addInterface(ObjectModelClassifier classifier,
                                Class<?> interfaceQualifiedName) {
        builder.addInterface(classifier, interfaceQualifiedName.getName());
    }

    protected void addImport(ObjectModelClassifier classifier,
                             String imports) {
        builder.addImport(classifier, imports);
    }

    protected void addImport(ObjectModelClassifier classifier,
                             ObjectModelClass imports) {
        builder.addImport(classifier, imports.getPackageName() + "." + imports.getName());
    }

    protected void addImport(ObjectModelClassifier classifier,
                             Class<?> imports) {
        builder.addImport(classifier, imports.getName());
    }

    public String getConstantName(String propertyName) {
        return getConstantPrefix() + builder.getConstantName(propertyName);
    }

    protected String getConstantPrefix(ObjectModelClassifier input,
                                       String defaultPrefix) {

        // look if there is a constant_prefix tag value
        String prefix = JavaGeneratorUtil.findTagValue(
                JavaGeneratorUtil.TAG_CONSTANT_PREFIX,
                input,
                getModel()
        );

        if (StringUtils.isEmpty(prefix)) {

            // no specific prefix, use default prefix
            prefix = defaultPrefix;
        }
        return prefix.trim();
    }
    
    protected ObjectModelAttribute addConstant(
            ObjectModelClassifier classifier,
            String name,
            String type,
            String value,
            ObjectModelModifier visibility) throws IllegalArgumentException {
        return builder.addConstant(classifier, name, type, value, visibility);
    }

    protected ObjectModelAttribute addConstant(
            ObjectModelClassifier classifier,
            String name,
            Class<?> type,
            String value,
            ObjectModelModifier visibility) throws IllegalArgumentException {
        return builder.addConstant(classifier, name, type.getName(), value, visibility);
    }

    protected ObjectModelAttribute addAttribute(
            ObjectModelClassifier classifier,
            String name,
            String type,
            String value,
            ObjectModelModifier... modifiers) {
        return builder.addAttribute(classifier, name, type, value, modifiers);
    }

    protected ObjectModelAttribute addAttribute(
            ObjectModelClassifier classifier,
            String name,
            Class<?> type,
            String value,
            ObjectModelModifier... modifiers) {
        return builder.addAttribute(classifier, name, type.getName(), value, modifiers);
    }

    protected ObjectModelAttribute addAttribute(
            ObjectModelClassifier classifier,
            ObjectModelAttribute attribute) {
        return builder.addAttribute(classifier, attribute);
    }

    public ObjectModelAttribute addAttribute(
            ObjectModelClassifier classifier,
            String name,
            String type) {
        return builder.addAttribute(classifier, name, type);
    }

    protected ObjectModelOperation addConstructor(
            ObjectModelClass clazz, ObjectModelModifier visibility)
            throws IllegalArgumentException {
        return builder.addConstructor(clazz, visibility);
    }

    protected ObjectModelOperation addConstructor(
            ObjectModelEnumeration clazz, ObjectModelModifier visibility)
            throws IllegalArgumentException {
        return builder.addConstructor(clazz, visibility);
    }

    protected ObjectModelOperation addOperation(
            ObjectModelClassifier classifier,
            String name,
            String type,
            ObjectModelModifier... modifiers) {
        return builder.addOperation(classifier, name, type, modifiers);
    }

    protected ObjectModelOperation addOperation(
            ObjectModelClassifier classifier,
            String name,
            Class<?> type,
            ObjectModelModifier... modifiers) {
        return builder.addOperation(classifier, name, type == null ? null : type.getName(), modifiers);
    }

    protected void addLiteral(ObjectModelEnumeration classifier, String name) {
        builder.addLiteral(classifier, name);
    }

    protected ObjectModelParameter addParameter(
            ObjectModelOperation operation,
            String type,
            String name) {
        return builder.addParameter(operation, type, name);
    }

    protected ObjectModelParameter addParameter(
            ObjectModelOperation operation,
            Class<?> type,
            String name) {
        return builder.addParameter(operation, type.getName(), name);
    }

    protected void addException(
            ObjectModelOperation operation,
            String exception) {
        builder.addException(operation, exception);
    }

    protected void addException(
            ObjectModelOperation operation,
            Class<?> exception) {
        builder.addException(operation, exception.getName());
    }

    protected void setOperationBody(ObjectModelOperation operation,
                                    String body)
            throws IllegalArgumentException {
        builder.setOperationBody(operation, body);
    }

    public ObjectModelClassifier addInnerClassifier(
            ObjectModelClass clazz,
            ObjectModelType type,
            String name) throws IllegalArgumentException {
        return builder.addInnerClassifier(clazz, type, name);
    }

    public void setDocumentation(ObjectModelElement element,
                                 String documentation) {
        builder.setDocumentation(element, documentation);
    }

    public void addAnnotation(ObjectModelClassifier classifier,
                              ObjectModelElement element,
                              String annotation) {
        builder.addAnnotation(classifier, element, annotation);
    }

    public ObjectModelOperation addBlock(ObjectModelClassifier classifier,
                                         ObjectModelModifier... modifiers) {
        return builder.addBlock(classifier, modifiers);
    }

    /**
     * Search in {@code input} model, for a enumeration dependency named
     * {@link JavaGeneratorUtil#DEPENDENCIES_CONSTANTS}.
     * <p/>
     * If found, add all literals on {@code output} model as constants.
     * <p/>
     * You can add a {@link JavaGeneratorUtil#TAG_CONSTANT_PREFIX} to your
     * enumeration to prefix constant names.
     *
     * @param input  the input model
     * @param output the ouput model
     * @return the set of constants names generated
     */
    protected Set<String> addConstantsFromDependency(ObjectModelClassifier input,
                                                     ObjectModelClassifier output) {

        // constant attributes via 'constants' enumeration dependency
        // literals of enumeration will be values and constant name will be generated
        ObjectModelDependency constantsDependencies =
                input.getDependency(JavaGeneratorUtil.DEPENDENCIES_CONSTANTS);

        if (constantsDependencies == null) {

            // no dependency found
            return Collections.emptySet();
        }

        ObjectModelClassifier classifier = constantsDependencies.getSupplier();

        if (!(classifier instanceof ObjectModelEnumeration)) {

            // do not have an enumeration
            return Collections.emptySet();
        }

        Set<String> constantNames = new HashSet<String>();
        ObjectModelEnumeration constants =
                (ObjectModelEnumeration) classifier;

        // add all literal of dependency as constants in interface
        Collection<String> literals = constants.getLiterals();
        for (String literal : literals) {

            // always use builder method to obtain constant name to not use a
            // possible altered by transformer
            String constantName = getConstantName(literal);

            addConstant(output,
                        constantName,
                        String.class,
                        "\"" + literal + "\"",
                        ObjectModelModifier.PUBLIC
            );
            constantNames.add(constantName);
        }

        return constantNames;
    }

    public String getConstantPrefix() {
        return constantPrefix;
    }

    public void setConstantPrefix(String constantPrefix) {
        this.constantPrefix = constantPrefix;
    }

    protected URL getFileInClassPath(String fqn) {
        String resourceName =  fqn.replaceAll("\\.", "/") + ".java";
        URL fileLocation = getClassLoader().getResource(resourceName);
        if (log.isDebugEnabled()) {
            log.debug("Look for resource : " + resourceName + " = " + fileLocation);
        }
        return fileLocation;
    }

}
